==> xkaapi-2.1.tar.gz: xkaapi-2.1/src/sched/daws/kaapi_sched_daws.c <==

#include <stdio.h>

#include <float.h>
#include <math.h>

#include "kaapi_impl.h"
#include "perfmodel/kaapi_perfmodel.h"

#define DAWS_MAX_ACTIVATED_TD    256

typedef struct kaapi_sched_daws_task_t {
  kaapi_taskdescr_t*        td;
  double                    speedup;
  double                    tcpu;
  double                    tgpu;
} kaapi_sched_daws_task_t;

/* executime time factor */
static double kaapi_sched_daws_alpha = 1.0f;
/* communication factor */
static double kaapi_sched_daws_beta = 1.0f;

static inline int kaapi_sched_daws_pushready_td(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push( rtl, td );
  }

  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);
  }
  
#if defined(KAAPI_USE_CUDA)
  /* optimize CUDA tasks if we are in a GPU thread (KAAPI_PROC_TYPE_CUDA) and
   it does not have a CPU body (KAAPI_PROC_TYPE_CPU). This line optmize DGEMM almost 2x.
   */
  if (kaapi_processor_get_type(curr_kproc) == KAAPI_PROC_TYPE_CUDA)
  {
    if ( (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) == 0) ||
        (kaapi_task_has_arch(td->task, KAAPI_PROC_TYPE_CPU) == 0) )
    {
      return kaapi_readylist_push( curr_kproc->rtl, td );
    }
  }
#endif
  
  /* for test only: push on master list task with highest priority */
  if ((td->priority == KAAPI_TASKLIST_MAX_PRIORITY) && (curr_kproc->kid != 0))
    return kaapi_readylist_remote_push( rtl, td );
  
  return kaapi_readylist_push( rtl, td );
}

static inline int kaapi_sched_daws_push_from_activationlist(kaapi_readytasklist_t* const rtl, kaapi_activationlink_t*	head)
{
  kaapi_taskdescr_t* td;
  int retval =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      ++retval;
      kaapi_sched_daws_pushready_td(rtl, td);
    }
    head = head->next;
  }
  return retval;
}

static inline uint32_t kaapi_sched_daws_pushactivated(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  uint32_t cnt_pushed= 0;
  
  /* push in the front the activated tasks */
  if (!kaapi_activationlist_isempty(&td->u.acl.list))
    cnt_pushed = kaapi_sched_daws_push_from_activationlist( rtl, td->u.acl.list.front );
  else
    cnt_pushed = 0;
  
  /* do bcast after child execution (they can produce output data) */
  if (td->u.acl.bcast !=0) 
    cnt_pushed += kaapi_sched_daws_push_from_activationlist( rtl, td->u.acl.bcast->front );
  
  return cnt_pushed;
}

static inline double kaapi_daws_execution_cost(kaapi_processor_t* const kproc,
                                               kaapi_taskdescr_t* const td )
{
  return kaapi_perfmodel_task_predict(kproc, td);
}

static inline double kaapi_daws_communication_cost( kaapi_processor_t* const kproc,
                                                   kaapi_taskdescr_t* const td )
{
  return kaapi_perfmodel_comm_predict_transfer(kproc, td);
}

static int kaapi_sched_task_cmp_cpu(const void *a, const void *b)
{
  return (((kaapi_sched_daws_task_t*)a)->speedup - ((kaapi_sched_daws_task_t*)b)->speedup);
}

static int kaapi_sched_task_cmp(const void *a, const void *b)
{
  return (((kaapi_sched_daws_task_t*)b)->speedup - ((kaapi_sched_daws_task_t*)a)->speedup);
}

static inline kaapi_taskdescr_t* kaapi_sched_daws_steal_by_affinity(kaapi_processor_t* const kproc,
                                                                          kaapi_sched_daws_task_t* const td_ready, const int n)
{
  int i;
  double predicted_end_min = DBL_MAX;
  double predicted_exectime, predicted_transfer, predicted_end;
  kaapi_sched_daws_task_t* td_ready_min;
  const int arch = kaapi_processor_get_type(kproc);
  
//  if(arch == KAAPI_PROC_TYPE_CPU)
//    qsort(td_ready, n, sizeof(kaapi_sched_daws_task_t), kaapi_sched_task_cmp_cpu);
//  else
//    qsort(td_ready, n, sizeof(kaapi_sched_daws_task_t), kaapi_sched_task_cmp);
  
  i = 0;
  while(td_ready[i].td != 0)
  {
    predicted_exectime = kaapi_daws_execution_cost(kproc, td_ready[i].td);
    predicted_transfer = kaapi_daws_communication_cost(kproc, td_ready[i].td);
    predicted_end = predicted_exectime * kaapi_sched_daws_alpha + predicted_transfer * kaapi_sched_daws_beta;
    if(predicted_end < predicted_end_min)
    {
      td_ready_min = &td_ready[i];
      predicted_end_min = predicted_end;
    }
    i++;
  }
  
#if 0
  if(td_ready_min != 0 && td_ready_min->td->fmt != 0)
  {
    fprintf(stdout, "%d:%s: td=%p (%s) speedup=%.4f STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td_ready_min->td, td_ready_min->td->fmt->name, td_ready_min->speedup, kproc->kid);
    fflush(stdout);
  }
#endif  
  
  return td_ready_min->td;
}

static inline void kaapi_sched_daws_compute_speedup(kaapi_sched_daws_task_t* const td_ready, kaapi_taskdescr_t* const td)
{
  td_ready->td = td;
  td_ready->tcpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_HOST, td);
  td_ready->tgpu = kaapi_perfmodel_task_predict_by_arch(KAAPI_PROC_TYPE_CUDA, td);
  td_ready->speedup = 0.f;
  if((fpclassify(td_ready->tcpu) != FP_ZERO) && (fpclassify(td_ready->tgpu) != FP_ZERO))
  {
    td_ready->speedup = td_ready->tcpu / td_ready->tgpu;
#if defined(DAWS_VERBOSE)
    fprintf(stdout,"%d:%s: td=%p(%s) speedup=%.4f\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td_ready->td, td_ready->td->fmt->name,
            td_ready->speedup);
    fflush(stdout);
#endif
  }
}

/* return the first td that can be executed on arch
 */
static inline kaapi_taskdescr_t* kaapi_sched_daws_steal_first(kaapi_processor_t* const thief, kaapi_taskdescr_t* const first)
{
  int arch = kaapi_processor_get_type(kaapi_get_current_processor());
  kaapi_taskdescr_t* td = first;
  kaapi_sched_daws_task_t td_ready[DAWS_MAX_ACTIVATED_TD];
  int i= 0;
  
  while(td != 0)
  {
    if( (td->fmt != 0) && (!kaapi_task_is_unstealable(td->task)) && kaapi_task_has_arch(td->task, arch) && kaapi_get_task_body_by_arch(td, arch))
    {
      kaapi_sched_daws_compute_speedup(&td_ready[i], td);
      i++;
    }
    td = td->prev;
  }
  
  if(i == 0)
  {
    return first;
  }
  
  td_ready[i].td = 0;
  if(i > 1)
  {
    return kaapi_sched_daws_steal_by_affinity(kaapi_get_current_processor(), td_ready, i);
  }
  
#if 0
  if(td != 0 && td->fmt != 0)
  {
    fprintf(stdout, "%d:%s: td=%p (%s) STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, thief->kid);
    fflush(stdout);
  }
#endif
  return first;
}

void kaapi_sched_daws_init(void)
{
  const char* alpha = getenv("KAAPI_DAWS_ALPHA");
  const char* beta = getenv("KAAPI_DAWS_BETA");
  
  if(alpha != 0)
  {
    kaapi_sched_daws_alpha = atof(alpha);
  }
  if(beta != 0)
  {
    kaapi_sched_daws_beta = atof(beta);
  }
  
  kaapi_default_param.perfmodel = 1;
  kaapi_perfmodel_init();  
}

void kaapi_sched_daws_finalize(void)
{
  if(kaapi_default_param.perfmodel)
  {
    kaapi_perfmodel_finalize();
  }
  kaapi_default_param.perfmodel = 0;  
}

uint32_t kaapi_sched_daws_push_activated( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  return kaapi_sched_daws_pushactivated( &tasklist->rtl, td );
}

void kaapi_sched_daws_push( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  kaapi_sched_daws_pushready_td(&tasklist->rtl, td);
}

kaapi_taskdescr_t* kaapi_sched_daws_steal(kaapi_processor_t* const thief, kaapi_taskdescr_t* const td)
{
  return kaapi_sched_daws_steal_first(thief, td);
}

void kaapi_sched_daws_task_epilogue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_epilogue(kaapi_get_current_processor(), td);
}

void kaapi_sched_daws_task_prologue(kaapi_taskdescr_t* const td)
{
  kaapi_perfmodel_prologue(kaapi_get_current_processor(), td);
}

==> xkaapi-2.2.tar.gz: xkaapi-2.2/src/sched/daws/kaapi_sched_daws.c <==

#include <stdio.h>

#include <float.h>
#include <math.h>

#include "kaapi_impl.h"

static inline int kaapi_sched_daws_pushready_td(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  kaapi_processor_t* curr_kproc = kaapi_get_current_processor();
  
  if(td->fmt == 0)
  {
    return kaapi_readylist_push(rtl, td);
  }
  
  if (kaapi_taskdescr_get_site(td) !=0)
  {
    return kaapi_sched_push_by_site(td);
  }
  
#if defined(KAAPI_USE_CUDA)
  /* TG: why to not do that for any kind of processor ?
   */
  if (kaapi_processor_get_type(curr_kproc) == KAAPI_PROC_TYPE_CUDA)
  {
    kaapi_processor_t* kproc_remote = kaapi_memory_taskdescr_affinity_find_reduce_transfer(curr_kproc, td);
    if ( kproc_remote != curr_kproc )
    {
      return kaapi_readylist_remote_push( kproc_remote->rtl, td );
    }
    
    if ( (kaapi_get_task_bodywh_by_arch(td, KAAPI_PROC_TYPE_CPU) == 0) ||
        (kaapi_task_has_arch(td->task, KAAPI_PROC_TYPE_CPU) == 0) )
    {
      return kaapi_readylist_push( curr_kproc->rtl, td );
    }
  }
#endif
  
  /* if td does not have GPU implementation, push it on master tasklist */
  if ( (kaapi_get_task_bodywh_by_arch(td, kaapi_processor_get_type(curr_kproc)) == 0) ||
      (kaapi_task_has_arch(td->task, kaapi_processor_get_type(curr_kproc)) == 0) )
  {
    kaapi_assert_debug( td->tasklist != 0 );
    return kaapi_readylist_push( &td->tasklist->rtl, td );
  }
  
  return kaapi_readylist_push( rtl, td );
}

static inline int kaapi_sched_daws_push_from_activationlist(kaapi_readytasklist_t* const rtl, kaapi_activationlink_t*	head)
{
  kaapi_taskdescr_t* td;
  int retval =0;
  
  while (head !=0) {
    td = head->td;
    if (kaapi_taskdescr_activated(td)) {
      ++retval;
      kaapi_sched_daws_pushready_td(rtl, td);
    }
    head = head->next;
  }
  return retval;
}

static inline uint32_t kaapi_sched_daws_pushactivated(kaapi_readytasklist_t* const rtl, kaapi_taskdescr_t* const td)
{
  uint32_t cnt_pushed= 0;
  
  /* push in the front the activated tasks */
  if (!kaapi_activationlist_isempty(&td->u.acl.list))
    cnt_pushed = kaapi_sched_daws_push_from_activationlist( rtl, td->u.acl.list.front );
  else
    cnt_pushed = 0;
  
  /* do bcast after child execution (they can produce output data) */
  if (td->u.acl.bcast !=0)
    cnt_pushed += kaapi_sched_daws_push_from_activationlist( rtl, td->u.acl.bcast->front );
  
  return cnt_pushed;
}

/* return the first td that can be executed on arch
 */
static inline kaapi_taskdescr_t* kaapi_sched_daws_steal_first(kaapi_processor_t* const thief, kaapi_taskdescr_t* const first)
{
  int arch = kaapi_processor_get_type(kaapi_get_current_processor());
  kaapi_taskdescr_t* td = first;
  //  int arch = kaapi_processor_get_type(thief);
  
  /* only steal for the righ processor arch or if fmt ==0 (means internal task) */
  /* TODO
   (kaapi_task_getbody(td->task) != (kaapi_task_body_t)kaapi_staticschedtask_body)
   */
  while ((td != 0) &&
         (kaapi_task_is_unstealable(td->task) || !kaapi_task_has_arch(td->task, arch) || ((td->fmt != 0) && (kaapi_get_task_body_by_arch(td, arch) == 0)))
         )
  {
    td = td->prev;
  }
#if 0
  if(td != 0 && td->fmt != 0)
  {
    fprintf(stdout, "%d:%s: td=%p (%s) STEAL (thief=%d)\n", kaapi_get_self_kid(), __FUNCTION__,
            (void*)td, td->fmt->name, thief->kid);
    fflush(stdout);
  }
#endif
  return td;
}

void kaapi_sched_daws_init(void)
{
}

void kaapi_sched_daws_finalize(void)
{
}

uint32_t kaapi_sched_daws_push_activated( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  return kaapi_sched_daws_pushactivated( &tasklist->rtl, td );
}

void kaapi_sched_daws_push( kaapi_tasklist_t*	const tasklist, kaapi_taskdescr_t* const td )
{
  kaapi_sched_daws_pushready_td(&tasklist->rtl, td);
}

int kaapi_sched_daws_pop( kaapi_readytasklist_t* rtl, kaapi_taskdescr_t** td )
{
  return kaapi_readylist_pop(rtl, td);
}

kaapi_taskdescr_t* kaapi_sched_daws_steal(kaapi_processor_t* const thief, kaapi_taskdescr_t* const td)
{
  return kaapi_sched_daws_steal_first(thief, td);
}

void kaapi_sched_daws_task_epilogue(kaapi_taskdescr_t* const td)
{
}

void kaapi_sched_daws_task_prologue(kaapi_taskdescr_t* const td)
{
}
